//! Physical constants for intensity measure calculation.

/// Gravitational acceleration in m/s².
pub const G: f64 = 9.81;

/// Gravitational acceleration in cm/s² (Gal).
pub const GAL: f64 = 981.0;
